
# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class RankingMode:
	GLOBAL_AROUND_SELF = 1
	GLOBAL = 2
	FRIENDS = 3


class Ranking2CategorySetting(common.Structure):
	def __init__(self):
		super().__init__()
		self.min_score = None
		self.max_score = None
		self.lowest_rank = None
		self.reset_month = None
		self.reset_day = None
		self.reset_hour = None
		self.reset_mode = None
		self.max_seasons_to_go_back = None
		self.score_order = None
	
	def check_required(self, settings, version):
		for field in ['min_score', 'max_score', 'lowest_rank', 'reset_month', 'reset_day', 'reset_hour', 'reset_mode', 'max_seasons_to_go_back', 'score_order']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.min_score = stream.u32()
		self.max_score = stream.u32()
		self.lowest_rank = stream.u32()
		self.reset_month = stream.u16()
		self.reset_day = stream.u8()
		self.reset_hour = stream.u8()
		self.reset_mode = stream.u8()
		self.max_seasons_to_go_back = stream.u8()
		self.score_order = stream.bool()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.min_score)
		stream.u32(self.max_score)
		stream.u32(self.lowest_rank)
		stream.u16(self.reset_month)
		stream.u8(self.reset_day)
		stream.u8(self.reset_hour)
		stream.u8(self.reset_mode)
		stream.u8(self.max_seasons_to_go_back)
		stream.bool(self.score_order)


class Ranking2ChartInfo(common.Structure):
	def __init__(self):
		super().__init__()
		self.create_time = None
		self.index = None
		self.category = None
		self.season = None
		self.bins_size = None
		self.sampling_rate = None
		self.score_order = None
		self.estimate_length = None
		self.estimate_highest_score = None
		self.estimate_lowest_score = None
		self.estimate_median_score = None
		self.estimate_average_score = None
		self.highest_bins_score = None
		self.lowest_bins_score = None
		self.bins_width = None
		self.attribute1 = None
		self.attribute2 = None
		self.quantities = None
	
	def check_required(self, settings, version):
		for field in ['create_time', 'index', 'category', 'season', 'bins_size', 'sampling_rate', 'score_order', 'estimate_length', 'estimate_highest_score', 'estimate_lowest_score', 'estimate_median_score', 'estimate_average_score', 'highest_bins_score', 'lowest_bins_score', 'bins_width', 'attribute1', 'attribute2', 'quantities']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.create_time = stream.datetime()
		self.index = stream.u32()
		self.category = stream.u32()
		self.season = stream.s32()
		self.bins_size = stream.u8()
		self.sampling_rate = stream.u8()
		self.score_order = stream.bool()
		self.estimate_length = stream.u32()
		self.estimate_highest_score = stream.u32()
		self.estimate_lowest_score = stream.u32()
		self.estimate_median_score = stream.u32()
		self.estimate_average_score = stream.double()
		self.highest_bins_score = stream.u32()
		self.lowest_bins_score = stream.u32()
		self.bins_width = stream.u32()
		self.attribute1 = stream.u32()
		self.attribute2 = stream.u32()
		self.quantities = stream.list(stream.u32)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.datetime(self.create_time)
		stream.u32(self.index)
		stream.u32(self.category)
		stream.s32(self.season)
		stream.u8(self.bins_size)
		stream.u8(self.sampling_rate)
		stream.bool(self.score_order)
		stream.u32(self.estimate_length)
		stream.u32(self.estimate_highest_score)
		stream.u32(self.estimate_lowest_score)
		stream.u32(self.estimate_median_score)
		stream.double(self.estimate_average_score)
		stream.u32(self.highest_bins_score)
		stream.u32(self.lowest_bins_score)
		stream.u32(self.bins_width)
		stream.u32(self.attribute1)
		stream.u32(self.attribute2)
		stream.list(self.quantities, stream.u32)


class Ranking2ChartInfoInput(common.Structure):
	def __init__(self):
		super().__init__()
		self.chart_index = None
		self.seasons_to_go_back = None
	
	def check_required(self, settings, version):
		for field in ['chart_index', 'seasons_to_go_back']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.chart_index = stream.u32()
		self.seasons_to_go_back = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.chart_index)
		stream.u8(self.seasons_to_go_back)


class Ranking2CommonData(common.Structure):
	def __init__(self):
		super().__init__()
		self.username = None
		self.mii = None
		self.binary_data = None
	
	def check_required(self, settings, version):
		for field in ['username', 'mii', 'binary_data']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.username = stream.string()
		self.mii = stream.qbuffer()
		self.binary_data = stream.qbuffer()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.string(self.username)
		stream.qbuffer(self.mii)
		stream.qbuffer(self.binary_data)


class Ranking2EstimateScoreRankInput(common.Structure):
	def __init__(self):
		super().__init__()
		self.category = None
		self.seasons_to_go_back = None
		self.score = None
	
	def check_required(self, settings, version):
		for field in ['category', 'seasons_to_go_back', 'score']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.category = stream.u32()
		self.seasons_to_go_back = stream.u8()
		self.score = stream.u32()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.category)
		stream.u8(self.seasons_to_go_back)
		stream.u32(self.score)


class Ranking2EstimateScoreRankOutput(common.Structure):
	def __init__(self):
		super().__init__()
		self.rank = None
		self.length = None
		self.score = None
		self.category = None
		self.season = None
		self.sampling_rate = None
	
	def check_required(self, settings, version):
		for field in ['rank', 'length', 'score', 'category', 'season', 'sampling_rate']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.rank = stream.u32()
		self.length = stream.u32()
		self.score = stream.u32()
		self.category = stream.u32()
		self.season = stream.s32()
		self.sampling_rate = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.rank)
		stream.u32(self.length)
		stream.u32(self.score)
		stream.u32(self.category)
		stream.s32(self.season)
		stream.u8(self.sampling_rate)


class Ranking2GetByListParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.category = None
		self.offset = None
		self.length = None
		self.sort_flags = None
		self.option_flags = None
		self.seasons_to_go_back = None
	
	def check_required(self, settings, version):
		for field in ['category', 'offset', 'length', 'sort_flags', 'option_flags', 'seasons_to_go_back']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.category = stream.u32()
		self.offset = stream.u32()
		self.length = stream.u32()
		self.sort_flags = stream.u32()
		self.option_flags = stream.u32()
		self.seasons_to_go_back = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.category)
		stream.u32(self.offset)
		stream.u32(self.length)
		stream.u32(self.sort_flags)
		stream.u32(self.option_flags)
		stream.u8(self.seasons_to_go_back)


class Ranking2GetParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.unique_id = 0
		self.pid = 0
		self.category = None
		self.offset = 0
		self.count = 10
		self.sort_flags = 0
		self.option_flags = 0
		self.mode = 2
		self.seasons_to_go_back = 0
	
	def check_required(self, settings, version):
		for field in ['category']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.unique_id = stream.u64()
		self.pid = stream.pid()
		self.category = stream.u32()
		self.offset = stream.u32()
		self.count = stream.u32()
		self.sort_flags = stream.u32()
		self.option_flags = stream.u32()
		self.mode = stream.u8()
		self.seasons_to_go_back = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u64(self.unique_id)
		stream.pid(self.pid)
		stream.u32(self.category)
		stream.u32(self.offset)
		stream.u32(self.count)
		stream.u32(self.sort_flags)
		stream.u32(self.option_flags)
		stream.u8(self.mode)
		stream.u8(self.seasons_to_go_back)


class Ranking2Info(common.Structure):
	def __init__(self):
		super().__init__()
		self.data = None
		self.lowest_rank = None
		self.num_entries = None
		self.season = None
	
	def check_required(self, settings, version):
		for field in ['data', 'lowest_rank', 'num_entries', 'season']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.data = stream.list(Ranking2RankData)
		self.lowest_rank = stream.u32()
		self.num_entries = stream.u32()
		self.season = stream.s32()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.list(self.data, stream.add)
		stream.u32(self.lowest_rank)
		stream.u32(self.num_entries)
		stream.s32(self.season)


class Ranking2RankData(common.Structure):
	def __init__(self):
		super().__init__()
		self.misc = None
		self.unique_id = None
		self.pid = None
		self.rank = None
		self.score = None
		self.common_data = Ranking2CommonData()
	
	def check_required(self, settings, version):
		for field in ['misc', 'unique_id', 'pid', 'rank', 'score']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.misc = stream.u64()
		self.unique_id = stream.u64()
		self.pid = stream.pid()
		self.rank = stream.u32()
		self.score = stream.u32()
		self.common_data = stream.extract(Ranking2CommonData)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u64(self.misc)
		stream.u64(self.unique_id)
		stream.pid(self.pid)
		stream.u32(self.rank)
		stream.u32(self.score)
		stream.add(self.common_data)


class Ranking2ScoreData(common.Structure):
	def __init__(self):
		super().__init__()
		self.misc = None
		self.category = None
		self.score = None
	
	def check_required(self, settings, version):
		for field in ['misc', 'category', 'score']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.misc = stream.u64()
		self.category = stream.u32()
		self.score = stream.u32()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u64(self.misc)
		stream.u32(self.category)
		stream.u32(self.score)


class Ranking2EstimateMyScoreRankInput(common.Structure):
	def __init__(self):
		super().__init__()
		self.category = None
		self.seasons_to_go_back = None
	
	def check_required(self, settings, version):
		for field in ['category', 'seasons_to_go_back']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.category = stream.u32()
		self.seasons_to_go_back = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.category)
		stream.u8(self.seasons_to_go_back)


class Ranking2Protocol:
	METHOD_PUT_SCORE = 1
	METHOD_GET_COMMON_DATA = 2
	METHOD_PUT_COMMON_DATA = 3
	METHOD_DELETE_COMMON_DATA = 4
	METHOD_GET_RANKING = 5
	METHOD_GET_RANKING_BY_PRINCIPAL_ID = 6
	METHOD_GET_CATEGORY_SETTING = 7
	METHOD_GET_RANKING_CHART = 8
	METHOD_GET_RANKING_CHARTS = 9
	METHOD_GET_ESTIMATE_SCORE_RANK = 10
	METHOD_GET_ESTIMATE_MY_SCORE_RANK = 11
	
	PROTOCOL_ID = 0x7A


class Ranking2Client(Ranking2Protocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def put_score(self, socres, unique_id):
		logger.info("Ranking2Client.put_score()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(socres, stream.add)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_PUT_SCORE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.put_score -> done")
	
	async def get_common_data(self, option_flags, pid, unique_id):
		logger.info("Ranking2Client.get_common_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(option_flags)
		stream.pid(pid)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_COMMON_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		common_data = stream.extract(Ranking2CommonData)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_common_data -> done")
		return common_data
	
	async def put_common_data(self, data, unique_id):
		logger.info("Ranking2Client.put_common_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(data)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_PUT_COMMON_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.put_common_data -> done")
	
	async def delete_common_data(self, unique_id):
		logger.info("Ranking2Client.delete_common_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELETE_COMMON_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.delete_common_data -> done")
	
	async def get_ranking(self, param):
		logger.info("Ranking2Client.get_ranking()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(param)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RANKING, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.extract(Ranking2Info)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_ranking -> done")
		return info
	
	async def get_ranking_by_principal_id(self, param, pids):
		logger.info("Ranking2Client.get_ranking_by_principal_id()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(param)
		stream.list(pids, stream.pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RANKING_BY_PRINCIPAL_ID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.extract(Ranking2Info)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_ranking_by_principal_id -> done")
		return info
	
	async def get_category_setting(self, category):
		logger.info("Ranking2Client.get_category_setting()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(category)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_CATEGORY_SETTING, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		setting = stream.extract(Ranking2CategorySetting)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_category_setting -> done")
		return setting
	
	async def get_ranking_chart(self, input):
		logger.info("Ranking2Client.get_ranking_chart()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(input)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RANKING_CHART, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.extract(Ranking2ChartInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_ranking_chart -> done")
		return info
	
	async def get_ranking_charts(self, inputs):
		logger.info("Ranking2Client.get_ranking_charts()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(inputs, stream.add)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RANKING_CHARTS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		infos = stream.list(Ranking2ChartInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_ranking_charts -> done")
		return infos
	
	async def get_estimate_score_rank(self, input):
		logger.info("Ranking2Client.get_estimate_score_rank()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(input)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_ESTIMATE_SCORE_RANK, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		output = stream.extract(Ranking2EstimateScoreRankOutput)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_estimate_score_rank -> done")
		return output
	
	async def get_estimate_my_score_rank(self, input):
		logger.info("Ranking2Client.get_estimate_my_score_rank()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(input)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_ESTIMATE_MY_SCORE_RANK, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		output = stream.extract(Ranking2EstimateScoreRankOutput)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("Ranking2Client.get_estimate_my_score_rank -> done")
		return output


class Ranking2Server(Ranking2Protocol):
	def __init__(self):
		self.methods = {
			self.METHOD_PUT_SCORE: self.handle_put_score,
			self.METHOD_GET_COMMON_DATA: self.handle_get_common_data,
			self.METHOD_PUT_COMMON_DATA: self.handle_put_common_data,
			self.METHOD_DELETE_COMMON_DATA: self.handle_delete_common_data,
			self.METHOD_GET_RANKING: self.handle_get_ranking,
			self.METHOD_GET_RANKING_BY_PRINCIPAL_ID: self.handle_get_ranking_by_principal_id,
			self.METHOD_GET_CATEGORY_SETTING: self.handle_get_category_setting,
			self.METHOD_GET_RANKING_CHART: self.handle_get_ranking_chart,
			self.METHOD_GET_RANKING_CHARTS: self.handle_get_ranking_charts,
			self.METHOD_GET_ESTIMATE_SCORE_RANK: self.handle_get_estimate_score_rank,
			self.METHOD_GET_ESTIMATE_MY_SCORE_RANK: self.handle_get_estimate_my_score_rank,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on Ranking2Server: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_put_score(self, client, input, output):
		logger.info("Ranking2Server.put_score()")
		#--- request ---
		socres = input.list(Ranking2ScoreData)
		unique_id = input.u64()
		await self.put_score(client, socres, unique_id)
	
	async def handle_get_common_data(self, client, input, output):
		logger.info("Ranking2Server.get_common_data()")
		#--- request ---
		option_flags = input.u32()
		pid = input.pid()
		unique_id = input.u64()
		response = await self.get_common_data(client, option_flags, pid, unique_id)
		
		#--- response ---
		if not isinstance(response, Ranking2CommonData):
			raise RuntimeError("Expected Ranking2CommonData, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_put_common_data(self, client, input, output):
		logger.info("Ranking2Server.put_common_data()")
		#--- request ---
		data = input.extract(Ranking2CommonData)
		unique_id = input.u64()
		await self.put_common_data(client, data, unique_id)
	
	async def handle_delete_common_data(self, client, input, output):
		logger.info("Ranking2Server.delete_common_data()")
		#--- request ---
		unique_id = input.u64()
		await self.delete_common_data(client, unique_id)
	
	async def handle_get_ranking(self, client, input, output):
		logger.info("Ranking2Server.get_ranking()")
		#--- request ---
		param = input.extract(Ranking2GetParam)
		response = await self.get_ranking(client, param)
		
		#--- response ---
		if not isinstance(response, Ranking2Info):
			raise RuntimeError("Expected Ranking2Info, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_ranking_by_principal_id(self, client, input, output):
		logger.info("Ranking2Server.get_ranking_by_principal_id()")
		#--- request ---
		param = input.extract(Ranking2GetByListParam)
		pids = input.list(input.pid)
		response = await self.get_ranking_by_principal_id(client, param, pids)
		
		#--- response ---
		if not isinstance(response, Ranking2Info):
			raise RuntimeError("Expected Ranking2Info, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_category_setting(self, client, input, output):
		logger.info("Ranking2Server.get_category_setting()")
		#--- request ---
		category = input.u32()
		response = await self.get_category_setting(client, category)
		
		#--- response ---
		if not isinstance(response, Ranking2CategorySetting):
			raise RuntimeError("Expected Ranking2CategorySetting, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_ranking_chart(self, client, input, output):
		logger.info("Ranking2Server.get_ranking_chart()")
		#--- request ---
		input = input.extract(Ranking2ChartInfoInput)
		response = await self.get_ranking_chart(client, input)
		
		#--- response ---
		if not isinstance(response, Ranking2ChartInfo):
			raise RuntimeError("Expected Ranking2ChartInfo, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_ranking_charts(self, client, input, output):
		logger.info("Ranking2Server.get_ranking_charts()")
		#--- request ---
		inputs = input.list(Ranking2ChartInfoInput)
		response = await self.get_ranking_charts(client, inputs)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_get_estimate_score_rank(self, client, input, output):
		logger.info("Ranking2Server.get_estimate_score_rank()")
		#--- request ---
		input = input.extract(Ranking2EstimateScoreRankInput)
		response = await self.get_estimate_score_rank(client, input)
		
		#--- response ---
		if not isinstance(response, Ranking2EstimateScoreRankOutput):
			raise RuntimeError("Expected Ranking2EstimateScoreRankOutput, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_estimate_my_score_rank(self, client, input, output):
		logger.info("Ranking2Server.get_estimate_my_score_rank()")
		#--- request ---
		input = input.extract(Ranking2EstimateMyScoreRankInput)
		response = await self.get_estimate_my_score_rank(client, input)
		
		#--- response ---
		if not isinstance(response, Ranking2EstimateScoreRankOutput):
			raise RuntimeError("Expected Ranking2EstimateScoreRankOutput, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def put_score(self, *args):
		logger.warning("Ranking2Server.put_score not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_common_data(self, *args):
		logger.warning("Ranking2Server.get_common_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def put_common_data(self, *args):
		logger.warning("Ranking2Server.put_common_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def delete_common_data(self, *args):
		logger.warning("Ranking2Server.delete_common_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_ranking(self, *args):
		logger.warning("Ranking2Server.get_ranking not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_ranking_by_principal_id(self, *args):
		logger.warning("Ranking2Server.get_ranking_by_principal_id not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_category_setting(self, *args):
		logger.warning("Ranking2Server.get_category_setting not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_ranking_chart(self, *args):
		logger.warning("Ranking2Server.get_ranking_chart not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_ranking_charts(self, *args):
		logger.warning("Ranking2Server.get_ranking_charts not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_estimate_score_rank(self, *args):
		logger.warning("Ranking2Server.get_estimate_score_rank not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_estimate_my_score_rank(self, *args):
		logger.warning("Ranking2Server.get_estimate_my_score_rank not implemented")
		raise common.RMCError("Core::NotImplemented")





# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class MonitoringProtocol:
	METHOD_PING_DAEMON = 1
	METHOD_GET_CLUSTER_MEMBERS = 2
	
	PROTOCOL_ID = 0x13


class MonitoringClient(MonitoringProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def ping_daemon(self):
		logger.info("MonitoringClient.ping_daemon()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_PING_DAEMON, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MonitoringClient.ping_daemon -> done")
		return result
	
	async def get_cluster_members(self):
		logger.info("MonitoringClient.get_cluster_members()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_CLUSTER_MEMBERS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		members = stream.list(stream.string)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MonitoringClient.get_cluster_members -> done")
		return members


class MonitoringServer(MonitoringProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_PING_DAEMON: self.handle_ping_daemon,
			self.METHOD_GET_CLUSTER_MEMBERS: self.handle_get_cluster_members,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on MonitoringServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_ping_daemon(self, client, input, output):
		logger.info("MonitoringServer.ping_daemon()")
		#--- request ---
		response = await self.ping_daemon(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_get_cluster_members(self, client, input, output):
		logger.info("MonitoringServer.get_cluster_members()")
		#--- request ---
		response = await self.get_cluster_members(client)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.string)
	
	async def ping_daemon(self, *args):
		logger.warning("MonitoringServer.ping_daemon not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_cluster_members(self, *args):
		logger.warning("MonitoringServer.get_cluster_members not implemented")
		raise common.RMCError("Core::NotImplemented")





# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class RecipientType:
	PRINCIPAL = 1
	GATHERING = 2


class MessageRecipient(common.Structure):
	def __init__(self):
		super().__init__()
		self.type = None
		self.pid = None
		self.gid = None
	
	def check_required(self, settings, version):
		for field in ['type', 'pid', 'gid']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.type = stream.u32()
		self.pid = stream.pid()
		self.gid = stream.u32()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.type)
		stream.pid(self.pid)
		stream.u32(self.gid)


class UserMessage(common.Data):
	def __init__(self):
		super().__init__()
		self.id = None
		self.parent_id = None
		self.sender = None
		self.reception_time = None
		self.life_time = None
		self.flags = None
		self.subject = None
		self.sender_name = None
		self.recipient = MessageRecipient()
	
	def check_required(self, settings, version):
		for field in ['id', 'parent_id', 'sender', 'reception_time', 'life_time', 'flags', 'subject', 'sender_name']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.id = stream.u32()
		self.parent_id = stream.u32()
		self.sender = stream.pid()
		self.reception_time = stream.datetime()
		self.life_time = stream.u32()
		self.flags = stream.u32()
		self.subject = stream.string()
		self.sender_name = stream.string()
		self.recipient = stream.extract(MessageRecipient)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.id)
		stream.u32(self.parent_id)
		stream.pid(self.sender)
		stream.datetime(self.reception_time)
		stream.u32(self.life_time)
		stream.u32(self.flags)
		stream.string(self.subject)
		stream.string(self.sender_name)
		stream.add(self.recipient)
common.DataHolder.register(UserMessage, "UserMessage")


class TextMessage(UserMessage):
	def __init__(self):
		super().__init__()
		self.body = None
	
	def check_required(self, settings, version):
		for field in ['body']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.body = stream.string()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.string(self.body)
common.DataHolder.register(TextMessage, "TextMessage")


class BinaryMessage(UserMessage):
	def __init__(self):
		super().__init__()
		self.body = None
	
	def check_required(self, settings, version):
		for field in ['body']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.body = stream.qbuffer()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.qbuffer(self.body)
common.DataHolder.register(BinaryMessage, "BinaryMessage")


class MessagingProtocol:
	METHOD_DELIVER_MESSAGE = 1
	METHOD_GET_NUMBER_OF_MESSAGES = 2
	METHOD_GET_MESSAGE_HEADERS = 3
	METHOD_RETRIEVE_ALL_MESSAGES_WITHIN_RANGE = 4
	METHOD_RETRIEVE_MESSAGES = 5
	METHOD_DELETE_MESSAGES = 6
	METHOD_DELETE_ALL_MESSAGES = 7
	
	PROTOCOL_ID = 0x17


class MessageDeliveryProtocol:
	NORESPONSE = True
	
	METHOD_DELIVER_MESSAGE = 1
	
	PROTOCOL_ID = 0x1B


class MessagingClient(MessagingProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def deliver_message(self, message):
		logger.info("MessagingClient.deliver_message()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.anydata(message)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELIVER_MESSAGE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.modified_message = stream.anydata()
		obj.sandbox_node_ids = stream.list(stream.u32)
		obj.participants = stream.list(stream.pid)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MessagingClient.deliver_message -> done")
		return obj
	
	async def get_number_of_messages(self, recipient):
		logger.info("MessagingClient.get_number_of_messages()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(recipient)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_NUMBER_OF_MESSAGES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		number = stream.u32()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MessagingClient.get_number_of_messages -> done")
		return number
	
	async def get_message_headers(self, recipient, range):
		logger.info("MessagingClient.get_message_headers()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(recipient)
		stream.add(range)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_MESSAGE_HEADERS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		headers = stream.list(UserMessage)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MessagingClient.get_message_headers -> done")
		return headers
	
	async def retrieve_all_messages_within_range(self, recipient, range):
		logger.info("MessagingClient.retrieve_all_messages_within_range()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(recipient)
		stream.add(range)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_RETRIEVE_ALL_MESSAGES_WITHIN_RANGE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		messages = stream.list(stream.anydata)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MessagingClient.retrieve_all_messages_within_range -> done")
		return messages
	
	async def retrieve_messages(self, recipient, message_ids, leave_on_server):
		logger.info("MessagingClient.retrieve_messages()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(recipient)
		stream.list(message_ids, stream.u32)
		stream.bool(leave_on_server)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_RETRIEVE_MESSAGES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		messages = stream.list(stream.anydata)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MessagingClient.retrieve_messages -> done")
		return messages
	
	async def delete_messages(self, recipient, message_ids):
		logger.info("MessagingClient.delete_messages()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(recipient)
		stream.list(message_ids, stream.u32)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELETE_MESSAGES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MessagingClient.delete_messages -> done")
	
	async def delete_all_messages(self, recipient):
		logger.info("MessagingClient.delete_all_messages()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(recipient)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELETE_ALL_MESSAGES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		number_deleted = stream.u32()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("MessagingClient.delete_all_messages -> done")
		return number_deleted


class MessageDeliveryClient(MessageDeliveryProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def deliver_message(self, message):
		logger.info("MessageDeliveryClient.deliver_message()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.anydata(message)
		await self.client.request(self.PROTOCOL_ID, self.METHOD_DELIVER_MESSAGE, stream.get(), True)


class MessagingServer(MessagingProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_DELIVER_MESSAGE: self.handle_deliver_message,
			self.METHOD_GET_NUMBER_OF_MESSAGES: self.handle_get_number_of_messages,
			self.METHOD_GET_MESSAGE_HEADERS: self.handle_get_message_headers,
			self.METHOD_RETRIEVE_ALL_MESSAGES_WITHIN_RANGE: self.handle_retrieve_all_messages_within_range,
			self.METHOD_RETRIEVE_MESSAGES: self.handle_retrieve_messages,
			self.METHOD_DELETE_MESSAGES: self.handle_delete_messages,
			self.METHOD_DELETE_ALL_MESSAGES: self.handle_delete_all_messages,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on MessagingServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_deliver_message(self, client, input, output):
		logger.info("MessagingServer.deliver_message()")
		#--- request ---
		message = input.anydata()
		response = await self.deliver_message(client, message)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['modified_message', 'sandbox_node_ids', 'participants']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.anydata(response.modified_message)
		output.list(response.sandbox_node_ids, output.u32)
		output.list(response.participants, output.pid)
	
	async def handle_get_number_of_messages(self, client, input, output):
		logger.info("MessagingServer.get_number_of_messages()")
		#--- request ---
		recipient = input.extract(MessageRecipient)
		response = await self.get_number_of_messages(client, recipient)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.u32(response)
	
	async def handle_get_message_headers(self, client, input, output):
		logger.info("MessagingServer.get_message_headers()")
		#--- request ---
		recipient = input.extract(MessageRecipient)
		range = input.extract(common.ResultRange)
		response = await self.get_message_headers(client, recipient, range)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_retrieve_all_messages_within_range(self, client, input, output):
		logger.info("MessagingServer.retrieve_all_messages_within_range()")
		#--- request ---
		recipient = input.extract(MessageRecipient)
		range = input.extract(common.ResultRange)
		response = await self.retrieve_all_messages_within_range(client, recipient, range)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.anydata)
	
	async def handle_retrieve_messages(self, client, input, output):
		logger.info("MessagingServer.retrieve_messages()")
		#--- request ---
		recipient = input.extract(MessageRecipient)
		message_ids = input.list(input.u32)
		leave_on_server = input.bool()
		response = await self.retrieve_messages(client, recipient, message_ids, leave_on_server)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.anydata)
	
	async def handle_delete_messages(self, client, input, output):
		logger.info("MessagingServer.delete_messages()")
		#--- request ---
		recipient = input.extract(MessageRecipient)
		message_ids = input.list(input.u32)
		await self.delete_messages(client, recipient, message_ids)
	
	async def handle_delete_all_messages(self, client, input, output):
		logger.info("MessagingServer.delete_all_messages()")
		#--- request ---
		recipient = input.extract(MessageRecipient)
		response = await self.delete_all_messages(client, recipient)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.u32(response)
	
	async def deliver_message(self, *args):
		logger.warning("MessagingServer.deliver_message not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_number_of_messages(self, *args):
		logger.warning("MessagingServer.get_number_of_messages not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_message_headers(self, *args):
		logger.warning("MessagingServer.get_message_headers not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def retrieve_all_messages_within_range(self, *args):
		logger.warning("MessagingServer.retrieve_all_messages_within_range not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def retrieve_messages(self, *args):
		logger.warning("MessagingServer.retrieve_messages not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def delete_messages(self, *args):
		logger.warning("MessagingServer.delete_messages not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def delete_all_messages(self, *args):
		logger.warning("MessagingServer.delete_all_messages not implemented")
		raise common.RMCError("Core::NotImplemented")


class MessageDeliveryServer(MessageDeliveryProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_DELIVER_MESSAGE: self.handle_deliver_message,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on MessageDeliveryServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_deliver_message(self, client, input, output):
		logger.info("MessageDeliveryServer.deliver_message()")
		#--- request ---
		message = input.anydata()
		await self.deliver_message(client, message)
	
	async def deliver_message(self, *args):
		logger.warning("MessageDeliveryServer.deliver_message not implemented")
		raise common.RMCError("Core::NotImplemented")





from anynet import tls, http
from nintendo import resources
from nintendo.switch import common, dauth
import base64

import logging
logger = logging.getLogger(__name__)


USER_AGENT = "NintendoSDK Firmware/%s (platform:NX; did:%016x; eid:lp1)"

API_VERSION = {
	 900: 1,
	 901: 1,
	 910: 1,
	 920: 1,
	1000: 1,
	1001: 1,
	1002: 1,
	1003: 1,
	1004: 1,
	1010: 1,
	1011: 1,
	1020: 1,
	1100: 1,
	1101: 1,
	1200: 1,
	1201: 1,
	1202: 1,
	1203: 1,
	1210: 1,
	1300: 1,
	1310: 1,
	1320: 1,
	1321: 1,
	1400: 1,
	1410: 1,
	1411: 1,
	1412: 1,
	1500: 1,
	1501: 1,
	1600: 1,
	1601: 1,
	1602: 1,
	1603: 1,
	1610: 1,
	1700: 1,
	1701: 1,
	1800: 1,
	1801: 1,
	1810: 1,
	1900: 1,
	1901: 1,
	2000: 2,
	2001: 2,
	2010: 2,
	2011: 2,
	2015: 2,
	2020: 2,
	2030: 2,
	2040: 2,
	2050: 2,
	2100: 2,
	2101: 2,
	2110: 2,
	2120: 2,
}

LATEST_VERSION = 2120


class DragonsError(Exception):
	def __init__(self, response):
		self.response = response
		
		self.type = response.json["type"]
		self.name = response.json["type"].split("/")[-1]
		self.title = response.json["title"]
		self.detail = response.json["detail"]
		self.status = response.json["number"]
		self.invalid_params = response.json.get("invalid-params")
	
	def __str__(self):
		return self.title


class DragonsClient:
	def __init__(self, device_id=None):
		self.device_id = device_id
		
		self.request_callback = http.request
		
		ca = resources.certificate("Nintendo_Class_2_CA_G3.der")
		self.context = tls.TLSContext()
		self.context.set_authority(ca)
		
		self.host_dragons = "dragons.hac.lp1.dragons.nintendo.net"
		self.host_dragonst = "dragonst.hac.lp1.dragons.nintendo.net"
		self.host_tigers = "tigers.hac.lp1.dragons.nintendo.net"
		
		self.system_version = LATEST_VERSION
		self.user_agent_nim = None
		if self.device_id is not None:
			self.user_agent_nim = USER_AGENT %(common.FIRMWARE_VERSIONS[LATEST_VERSION], self.device_id)
		self.user_agent_dauth = dauth.USER_AGENT[LATEST_VERSION]
		self.api_version = API_VERSION[self.system_version]
	
	def set_request_callback(self, callback): self.request_callback = callback
	
	def set_context(self, context): self.context = context
	def set_certificate(self, cert, key): self.context.set_certificate(cert, key)
	
	def set_hosts(self, dragons, dragonst, tigers):
		self.host_dragons = dragons
		self.host_dragonst = dragonst
		self.host_tigers = tigers
	
	def set_system_version(self, version):
		if version not in common.FIRMWARE_VERSIONS:
			raise ValueError("Unknown system version: %i" %version)
		
		self.system_version = version
		if self.device_id is not None:
			self.user_agent_nim = USER_AGENT %(common.FIRMWARE_VERSIONS[version], self.device_id)
		self.user_agent_dauth = dauth.USER_AGENT[version]
		self.api_version = API_VERSION[self.system_version]
	
	async def request(self, req, host, device_token, *, account_id=None):
		if self.user_agent_nim is None:
			raise ValueError("This request requires a device id")
		
		req.headers["Host"] = host
		req.headers["Accept"] = "*/*"
		req.headers["User-Agent"] = self.user_agent_nim
		req.headers["DeviceAuthorization"] = "Bearer " + device_token
		if account_id is not None:
			req.headers["Nintendo-Account-Id"] = "%016x" %account_id
		if req.json is not None:
			req.headers["Content-Type"] = "application/json"
			req.headers["Content-Length"] = 0
		else:
			req.headers["Content-Length"] = 0
			req.headers["Content-Type"] = "application/x-www-form-urlencoded"
		
		response = await self.request_callback(host, req, self.context)
		if response.error() and response.json:
			logger.error("Dragons server returned an error: %s" %response.json)
			raise DragonsError(response)
		response.raise_if_error()
		return response
	
	async def request_dauth(self, req, device_token, title_id):
		req.headers["Host"] = self.host_dragons
		if self.system_version < 1800:
			req.headers["User-Agent"] = self.user_agent_dauth
		req.headers["Accept"] = "*/*"
		if self.system_version >= 2000:
			req.headers["User-Agent"] = self.user_agent_dauth
		req.headers["Content-Type"] = "application/json"
		req.headers["DeviceAuthorization"] = "Bearer " + device_token
		req.headers["Nintendo-Application-Id"] = "%016x" %title_id
		req.headers["Content-Length"] = 0
		
		response = await self.request_callback(self.host_dragons, req, self.context)
		if response.error() and response.json:
			logger.error("Dragons server returned an error: %s" %response.json)
			raise DragonsError(response)
		response.raise_if_error()
		return response
	
	async def publish_elicense_archive(self, device_token, challenge, certificate, account_id):
		req = http.HTTPRequest.post("/v%i/elicense_archives/publish" %self.api_version)
		req.json = {
			"challenge": challenge,
			"certificate": base64.b64encode(certificate).decode()
		}
		
		response = await self.request(req, self.host_dragons, device_token, account_id=account_id)
		return response.json
	
	async def report_elicense_archive(self, device_token, elicense_archive_id, account_id):
		req = http.HTTPRequest.put("/v%i/elicense_archives/%s/report" %(self.api_version, elicense_archive_id))
		await self.request(req, self.host_dragons, device_token, account_id=account_id)
	
	async def publish_device_linked_elicenses(self, device_token):
		req = http.HTTPRequest.post("/v%i/rights/publish_device_linked_elicenses" %self.api_version)
		response = await self.request(req, self.host_dragons, device_token)
		return response.json
	
	async def exercise_elicense(self, device_token, elicense_ids, account_ids, current_account_id):
		req = http.HTTPRequest.post("/v%i/elicenses/exercise" %self.api_version)
		req.json = {
			"elicense_ids": elicense_ids,
			"account_ids": ["%016x" %i for i in account_ids]
		}
		await self.request(req, self.host_dragons, device_token, account_id=current_account_id)
	
	async def contents_authorization_token_for_aauth(self, device_token, elicense_id, na_id, title_id):
		if self.system_version < 1500:
			raise ValueError("contents_authorization_token_for_aauth was added in system version 15.0.0")
		
		req = http.HTTPRequest.post("/v%i/contents_authorization_token_for_aauth/issue" %self.api_version)
		req.json = {
			"elicense_id": elicense_id,
			"na_id": "%016x" %na_id
		}
		
		response = await self.request_dauth(req, device_token, title_id)
		return response.json




from Crypto.Util.Padding import pad
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Hash import SHA256
from Crypto.Random import get_random_bytes
from anynet import tls, http
from nintendo import resources
import struct
import base64

import logging
logger = logging.getLogger(__name__)


RSA_MODULUS =  int(
	"2903599220185509629948246004681271806662185201109683699434876284"
	"9306378942456577312580648895443616535088601867223713942187399041"
	"4854872772034425863387471719375473684104853507768450203982749294"
	"8967945831738920699512732919046223059402955082098739033920491649"
	"6879108565068863591362496844602988110766097564477545097467537357"
	"3183749964356608012071405755940871989007021731074728723470759285"
	"8155278592486462922448753256166071381157786436864032235809243318"
	"2591089355065715995209202752330511548478896205428626608544326862"
	"4782505679727111312756904637828438785474471375120478991907979820"
	"98870089661523253199182993983393803812441"
)

RSA_EXPONENT = 65537


USER_AGENT = {
	 900: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 901: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 910: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 920: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	1000: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1001: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1002: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1003: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1004: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1010: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1011: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1020: "libcurl (nnAccount; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1100: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1101: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1200: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1201: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1202: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1203: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1210: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1300: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.3.0.0; Add-on 13.3.0.0)",
	1310: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1320: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1321: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1400: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1410: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1411: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1412: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1500: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 15.3.0.0; Add-on 15.3.0.0)",
	1501: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 15.3.0.0; Add-on 15.3.0.0)",
	1600: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1601: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1602: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1603: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1610: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1700: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 17.5.0.0; Add-on 17.5.0.0)",
	1701: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 17.5.0.0; Add-on 17.5.0.0)",
	1800: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1801: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1810: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1900: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 19.3.0.0; Add-on 19.3.0.0)",
	1901: "libcurl (nnHttp; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 19.3.0.0; Add-on 19.3.0.0)",
	2000: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2001: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2010: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2011: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2015: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2020: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2030: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2040: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2050: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2100: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
	2101: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
	2110: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
	2120: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
}

API_VERSION = {
	 900: 3,
	 901: 3,
	 910: 3,
	 920: 3,
	1000: 3,
	1001: 3,
	1002: 3,
	1003: 3,
	1004: 3,
	1010: 3,
	1011: 3,
	1020: 3,
	1100: 3,
	1101: 3,
	1200: 3,
	1201: 3,
	1202: 3,
	1203: 3,
	1210: 3,
	1300: 3,
	1310: 3,
	1320: 3,
	1321: 3,
	1400: 3,
	1410: 3,
	1411: 3,
	1412: 3,
	1500: 4,
	1501: 4,
	1600: 4,
	1601: 4,
	1602: 4,
	1603: 4,
	1610: 4,
	1700: 4,
	1701: 4,
	1800: 4,
	1801: 4,
	1810: 4,
	1900: 5,
	1901: 5,
	2000: 5,
	2001: 5,
	2010: 5,
	2011: 5,
	2015: 5,
	2020: 5,
	2030: 5,
	2040: 5,
	2050: 5,
	2100: 5,
	2101: 5,
	2110: 5,
	2120: 5,
}

LATEST_VERSION = 2120


class AAuthError(Exception):
	DEVICE_TOKEN_EXPIRED = 103
	ROMID_BANNED = 105
	UNAUTHORIZED_APPLICATION = 106
	SERVICE_CLOSED = 109
	APPLICATION_UPDATE_REQUIRED = 111
	INTERNAL_SERVER_ERROR = 112
	GENERIC = 118
	REGION_MISMATCH = 121
	
	def __init__(self, response):
		self.response = response
		self.code = int(response.json["errors"][0]["code"])
		self.message = response.json["errors"][0]["message"]
	
	def __str__(self):
		return self.message


class AAuthClient:
	def __init__(self):
		self.request_callback = http.request
		
		self.context = tls.TLSContext()
		self.context_overridden = False
		
		self.host_overridden = False

		self.power_state = "FA"

		self.set_system_version(LATEST_VERSION)

	def set_request_callback(self, callback): self.request_callback = callback

	def set_context(self, context):
		self.context = context
		self.context_overridden = True
	
	def set_certificate(self, cert, key):
		self.context.set_certificate(cert, key)
	
	def set_host(self, host):
		self.host = host
		self.host_overridden = True
	
	def set_power_state(self, state): self.power_state = state
	
	def set_system_version(self, version):
		if version not in USER_AGENT:
			raise ValueError("Unknown system version")
		
		self.system_version = version
		self.user_agent = USER_AGENT[version]
		self.api_version = API_VERSION[version]

		if version >= 2000:
			caname = "Nintendo_Root_CA_G4.der"
			host = "aauth.hac.lp1.ndas.srv.nintendo.net"
		else:
			caname = "Nintendo_CA_G3.der"
			host = "aauth-lp1.ndas.srv.nintendo.net"
		
		if not self.context_overridden:
			ca = resources.certificate(caname)
			self.context.set_authority(ca)
		
		if not self.host_overridden:
			self.host = host
	
	async def request(self, req, use_power_state):
		if self.system_version < 1800:
			req.headers["Host"] = self.host
			req.headers["User-Agent"] = self.user_agent
			req.headers["Accept"] = "*/*"
			if use_power_state:
				req.headers["X-Nintendo-PowerState"] = self.power_state
			req.headers["Content-Length"] = 0
			req.headers["Content-Type"] = "application/x-www-form-urlencoded"
		elif self.system_version < 2000:
			req.headers["Host"] = self.host
			req.headers["Accept"] = "*/*"
			req.headers["Content-Type"] = "application/x-www-form-urlencoded"	
			if use_power_state:
				req.headers["X-Nintendo-PowerState"] = self.power_state
			req.headers["Content-Length"] = 0
		else:
			req.headers["Host"] = self.host
			req.headers["Accept"] = "*/*"
			req.headers["User-Agent"] = self.user_agent
			req.headers["Content-Type"] = "application/x-www-form-urlencoded"
			if use_power_state:
				req.headers["X-Nintendo-PowerState"] = self.power_state
			req.headers["Content-Length"] = 0
		
		response = await self.request_callback(self.host, req, self.context)
		if response.json and "errors" in response.json:
			logger.error("AAuth server returned errors:")
			for error in response.json["errors"]:
				logger.error("  (%s) %s", error["code"], error["message"])
			raise AAuthError(response)
		response.raise_if_error()
		return response
	
	def verify_ticket(self, ticket, title_id):
		# Verify ticket, in case someone accidentally
		# provides the wrong ticket
		if len(ticket) != 0x2C0:
			raise ValueError("Ticket has unexpected size")
		if struct.unpack_from("<I", ticket)[0] != 0x10004:
			raise ValueError("Ticket has invalid signature type")
		if struct.unpack_from(">Q", ticket, 0x2A0)[0] != title_id:
			raise ValueError("Ticket has different title id")
		if struct.unpack_from(">Q", ticket, 0x2A8)[0] != ticket[0x285]:
			raise ValueError("Ticket has inconsistent master key revision")
	
	def verify_token(self, token):
		# Just a basic check, to make sure that people do not
		# accidentally pass a ticket instead of a JWT
		if not isinstance(token, str) or token.count(".") != 2:
			raise ValueError("Cert must contain a valid JWT")
	
	async def get_time(self):
		req = http.HTTPRequest.get("/v1/time")
		req.headers["Host"] = self.host
		req.headers["Accept"] = "*/*"
		
		response = await http.request(self.host, req, self.context)
		response.raise_if_error()
		
		time = int(response.headers["X-NINTENDO-UNIXTIME"])
		ip = response.headers["X-NINTENDO-GLOBAL-IP"]
		return time, ip
	
	async def challenge(self, device_token):
		param_name = "&device_auth_token" if self.system_version < 1800 else "device_auth_token"

		req = http.HTTPRequest.post("/v%i/challenge" %self.api_version)
		req.rawform = {
			param_name: device_token
		}
		
		response = await self.request(req, False)
		return response.json

	# Warning: do not use auth_nocert on a production server.
	# It will immediately ban your Switch.
	async def auth_nocert(self, title_id, title_version, device_token):
		auth_type = "media_type" if self.api_version < 5 else "auth_type"

		req = http.HTTPRequest.post("/v%i/application_auth_token" %self.api_version)
		req.form = {
			"application_id": "%016x" %title_id,
			"application_version": "%08x" %title_version,
			"device_auth_token": device_token,
			auth_type: "NO_CERT"
		}
		
		response = await self.request(req, True)
		return response.json

	async def auth_system(self, title_id, title_version, device_token):
		auth_type = "media_type" if self.api_version < 5 else "auth_type"

		req = http.HTTPRequest.post("/v%i/application_auth_token" %self.api_version)
		req.form = {
			"application_id": "%016x" %title_id,
			"application_version": "%08x" %title_version,
			"device_auth_token": device_token,
			auth_type: "SYSTEM"
		}

		response = await self.request(req, True)
		return response.json

	async def auth_digital(self, title_id, title_version, device_token, cert):
		auth_type = "media_type" if self.api_version < 5 else "auth_type"

		req = http.HTTPRequest.post("/v%i/application_auth_token" %self.api_version)
		req.form = {
			"application_id": "%016x" %title_id,
			"application_version": "%08x" %title_version,
			"device_auth_token": device_token,
			auth_type: "DIGITAL"
		}
		
		if self.api_version == 3:
			self.verify_ticket(cert, title_id)
			
			plain_key = get_random_bytes(16)
			
			aes = AES.new(plain_key, AES.MODE_CBC, iv=bytes(16))
			encrypted_ticket = aes.encrypt(pad(cert, 16))
			
			rsa_key = RSA.construct((RSA_MODULUS, RSA_EXPONENT))
			rsa = PKCS1_OAEP.new(rsa_key, SHA256)
			encrypted_key = rsa.encrypt(plain_key)
		
			req.form["cert"] = base64.b64encode(encrypted_ticket, b"-_").decode().rstrip("=")
			req.form["cert_key"] = base64.b64encode(encrypted_key, b"-_").decode().rstrip("=")
		
		elif self.api_version >= 4:
			self.verify_token(cert)
			
			req.form["cert"] = cert
		
		response = await self.request(req, True)
		return response.json

	async def auth_gamecard(self, title_id, title_version, device_token, cert, gvt, challenge=None, challenge_src=None):
		auth_type = "media_type" if self.api_version < 5 else "auth_type"

		req = http.HTTPRequest.post("/v%i/application_auth_token" %self.api_version)
		req.form = {
			"application_id": "%016x" %title_id,
			"application_version": "%08x" %title_version,
			"device_auth_token": device_token,
			auth_type: "GAMECARD",
			"gvt": base64.b64encode(gvt, b"-_").decode().rstrip("="),
			"cert": base64.b64encode(cert, b"-_").decode().rstrip("=")
		}

		if self.api_version >= 5:
			req.form["challenge"] = challenge
			req.form["challenge_src"] = challenge_src

		response = await self.request(req, True)
		return response.json




from anynet import util
import struct


PRIMARY_ADDRESS = "nncs1-lp1.n.n.srv.nintendo.net"
PRIMARY_PORT = 10025


async def detect_external_address(socket):
	message = struct.pack(">IIII", 1, 0, 0, 0)
	await socket.send(message, (PRIMARY_ADDRESS, PRIMARY_PORT))
	
	while True:
		response = (await socket.recv())[0]
		if len(response) == 16:
			type, port, host, extra = struct.unpack(">IIII", response)
			if type == 1:
				return util.ip_from_hex(host), port




from anynet import http, tls
from nintendo import resources
import datetime
import secrets
import base64

import logging
logger = logging.getLogger(__name__)


MEDIA_TYPE_SYSTEM = 0
MEDIA_TYPE_DIGITAL = 1
MEDIA_TYPE_CARTRIDGE = 2


def b64decode(text):
	text = text.replace(".", "+").replace("-", "/").replace("*", "=")
	return base64.b64decode(text)

def b64encode(text):
	# Convert to bytes if necessary
	if isinstance(text, str):
		text = text.encode()
	
	text = base64.b64encode(text).decode()
	return text.replace("+", ".").replace("/", "-").replace("=", "*")

def decode_form(form):
	return {key: b64decode(value) for key, value in form.items()}

def encode_form(form):
	return {key: b64encode(value) for key, value in form.items()}

def parse_date(text):
	return datetime.datetime.strptime(text, "%Y%m%d%H%M%S")


# Known error codes:
#	001: success
#   109: missing or malformed parameter in request
#   110: game server is no longer available
#   119: fpd version is outdated
#   121: device certificate is invalid
#   122: uid hmac is invalid
#   125: game id is invalid

class NASCError(Exception):
	def __init__(self, status_code, form):
		self.status_code = status_code
		
		returncd = form["returncd"].decode()
		self.return_code = None if returncd == "null" else int(returncd)
		self.retry = bool(int(form["retry"].decode()))
		self.datetime = parse_date(form["datetime"].decode())
	
	def __str__(self):
		if self.return_code is None:
			return "NASC request failed with error code null"
		return "NASC request failed with error code %i" %self.return_code


class LoginResponse:
	def __init__(self):
		self.host = None
		self.port = None
		self.token = None
		self.datetime = None

	@classmethod
	def parse(cls, form):
		host, port = form["locator"].decode().split(":")
		
		inst = cls()
		inst.host = host
		inst.port = int(port)
		inst.token = b64encode(form["token"])
		inst.datetime = parse_date(form["datetime"].decode())
		return inst


class NASCClient:
	def __init__(self):
		self.url = "nasc.nintendowifi.net"

		ca = resources.certificate("Nintendo_CA_G3.der")
		cert = resources.certificate("CTR_Common_Prod_1.der")
		key = resources.private_key("CTR_Common_Prod_1.key")

		self.context = tls.TLSContext()
		self.context.set_authority(ca)
		self.context.set_certificate(cert, key)
		
		self.sdk_version_major = 0
		self.sdk_version_minor = 0
		
		self.title_id = None
		self.title_version = None
		self.product_code = "----"
		self.maker_code = "00"
		self.media_type = MEDIA_TYPE_SYSTEM
		self.rom_id = None
		
		self.serial_number = None
		self.mac_address = None
		self.fcd_cert = None
		self.device_name = ""
		self.unit_code = "2"
		
		self.bss_id = secrets.token_hex(6)
		self.ap_info = "01:0000000000"
		
		self.region = 3
		self.language = 2
		
		self.pid = None
		self.pid_hmac = None
		self.password = None
		
		self.fpd_version = 16
		self.environment = "L1"
	
	def set_context(self, context):
		self.context = context
	
	def set_url(self, url): self.url = url
	
	def set_sdk_version(self, major_version, minor_version):
		self.sdk_major_version = major_version
		self.sdk_minor_version = minor_version
	
	def set_title(
		self, title_id, title_version, product_code="----", maker_code="00",
		media_type=MEDIA_TYPE_SYSTEM, rom_id=None
	):
		if media_type == MEDIA_TYPE_CARTRIDGE and rom_id is None:
			raise ValueError("Rom ID is required for cartridge media type")
		
		self.title_id = title_id
		self.title_version = title_version
		self.product_code = product_code
		self.maker_code = maker_code
		self.media_type = media_type
		self.rom_id = rom_id
	
	def set_device(self, serial_number, mac_address, fcd_cert, name="", unit_code="2"):
		self.serial_number = serial_number
		self.mac_address = mac_address
		self.fcd_cert = fcd_cert
		self.device_name = name
		self.unit_code = unit_code
	
	def set_network(self, bss_id, ap_info):
		self.bss_id = bss_id
		self.ap_info = ap_info
	
	def set_locale(self, region, language):
		self.region = region
		self.language = language
	
	def set_user(self, pid, pid_hmac):
		self.pid = pid
		self.pid_hmac = pid_hmac
		self.password = None
		
	def set_password(self, password):
		self.pid = None
		self.pid_hmac = None
		self.password = password
	
	def set_fpd_version(self, version): self.fpd_version = version
	def set_environment(self, environment): self.environment = environment

	async def request(self, req):
		# Apply Nintendo's custom base64 encoding
		req.form = encode_form(req.form)
		
		# Must manually decode form here since server doesn't return correct content-type
		response = await http.request(self.url, req, self.context)
		response.form = decode_form(http.formdecode(response.text))
		
		return_code = response.form["returncd"].decode()
		if return_code == "null" or int(return_code) != 1:
			raise NASCError(response.status_code, response.form)
		
		return response

	async def login(self, game_server_id, nickname=""):
		if self.title_id is None:
			raise ValueError("Please configure the title (set_title)")
		if self.serial_number is None:
			raise ValueError("Please configure the device (set_device)")
		if self.pid is None and self.password is None:
			raise ValueError("Please configure a user id or password (set_user / set_password)")
		
		req = http.HTTPRequest.post("/ac")
		req.headers["Host"] = self.url
		req.headers["X-GameId"] = "%08X" %game_server_id
		req.headers["User-Agent"] = "CTR FPD/%04X" %self.fpd_version
		
		# The real 3DS seems to add this header twice for some reason
		req.headers.add("Content-Type", "application/x-www-form-urlencoded")
		req.headers.add("Content-Type", "application/x-www-form-urlencoded")

		now = datetime.datetime.now()
		device_time = now.strftime("%y%m%d%H%M%S")

		req.form = {}
		req.form["gameid"] = "%08X" %game_server_id
		req.form["sdkver"] = "%03i%03i" %(self.sdk_version_major, self.sdk_version_minor)
		req.form["titleid"] = "%016X" %self.title_id
		req.form["gamecd"] = self.product_code
		req.form["gamever"] = "%04X" %self.title_version
		req.form["mediatype"] = str(self.media_type)
		if self.media_type == MEDIA_TYPE_CARTRIDGE:
			req.form["romid"] = self.rom_id
		req.form["makercd"] = self.maker_code
		req.form["unitcd"] = self.unit_code
		req.form["macadr"] = self.mac_address
		req.form["bssid"] = self.bss_id
		req.form["apinfo"] = self.ap_info
		req.form["fcdcert"] = self.fcd_cert
		req.form["devname"] = self.device_name.encode("utf-16-le")
		req.form["servertype"] = self.environment
		req.form["fpdver"] = "%04X" %self.fpd_version
		req.form["devtime"] = device_time
		req.form["lang"] = "%02X" %self.language
		req.form["region"] = "%02X" %self.region
		req.form["csnum"] = self.serial_number
		if self.pid_hmac is not None:
			req.form["uidhmac"] = self.pid_hmac
			req.form["userid"] = str(self.pid)
		else:
			req.form["passwd"] = self.password
		req.form["action"] = "LOGIN"
		req.form["ingamesn"] = nickname
		
		response = await self.request(req)
		return LoginResponse.parse(response.form)




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class ApiCall(common.Structure):
	def __init__(self):
		super().__init__()
		self.name = None
		self.time = None
		self.pid = None
	
	def check_required(self, settings, version):
		for field in ['name', 'time', 'pid']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.name = stream.string()
		self.time = stream.datetime()
		self.pid = stream.pid()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.string(self.name)
		stream.datetime(self.time)
		stream.pid(self.pid)


class DebugProtocol:
	METHOD_ENABLE_API_RECORDER = 1
	METHOD_DISABLE_API_RECORDER = 2
	METHOD_IS_API_RECORDER_ENABLED = 3
	METHOD_GET_API_CALLS = 4
	METHOD_SET_EXCLUDE_JOINED_MATCHMAKE_SESSION = 5
	METHOD_GET_EXCLUDE_JOINED_MATCHMAKE_SESSION = 6
	METHOD_GET_API_CALL_SUMMARY = 7
	
	PROTOCOL_ID = 0x74


class DebugClient(DebugProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def enable_api_recorder(self):
		logger.info("DebugClient.enable_api_recorder()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_ENABLE_API_RECORDER, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("DebugClient.enable_api_recorder -> done")
	
	async def disable_api_recorder(self):
		logger.info("DebugClient.disable_api_recorder()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DISABLE_API_RECORDER, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("DebugClient.disable_api_recorder -> done")
	
	async def is_api_recorder_enabled(self):
		logger.info("DebugClient.is_api_recorder_enabled()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_IS_API_RECORDER_ENABLED, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		enabled = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("DebugClient.is_api_recorder_enabled -> done")
		return enabled
	
	async def get_api_calls(self, pids, unk1, unk2):
		logger.info("DebugClient.get_api_calls()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(pids, stream.pid)
		stream.datetime(unk1)
		stream.datetime(unk2)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_API_CALLS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		calls = stream.list(ApiCall)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("DebugClient.get_api_calls -> done")
		return calls


class DebugServer(DebugProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_ENABLE_API_RECORDER: self.handle_enable_api_recorder,
			self.METHOD_DISABLE_API_RECORDER: self.handle_disable_api_recorder,
			self.METHOD_IS_API_RECORDER_ENABLED: self.handle_is_api_recorder_enabled,
			self.METHOD_GET_API_CALLS: self.handle_get_api_calls,
			self.METHOD_SET_EXCLUDE_JOINED_MATCHMAKE_SESSION: self.handle_set_exclude_joined_matchmake_session,
			self.METHOD_GET_EXCLUDE_JOINED_MATCHMAKE_SESSION: self.handle_get_exclude_joined_matchmake_session,
			self.METHOD_GET_API_CALL_SUMMARY: self.handle_get_api_call_summary,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on DebugServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_enable_api_recorder(self, client, input, output):
		logger.info("DebugServer.enable_api_recorder()")
		#--- request ---
		await self.enable_api_recorder(client)
	
	async def handle_disable_api_recorder(self, client, input, output):
		logger.info("DebugServer.disable_api_recorder()")
		#--- request ---
		await self.disable_api_recorder(client)
	
	async def handle_is_api_recorder_enabled(self, client, input, output):
		logger.info("DebugServer.is_api_recorder_enabled()")
		#--- request ---
		response = await self.is_api_recorder_enabled(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_get_api_calls(self, client, input, output):
		logger.info("DebugServer.get_api_calls()")
		#--- request ---
		pids = input.list(input.pid)
		unk1 = input.datetime()
		unk2 = input.datetime()
		response = await self.get_api_calls(client, pids, unk1, unk2)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_set_exclude_joined_matchmake_session(self, client, input, output):
		logger.warning("DebugServer.set_exclude_joined_matchmake_session is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_exclude_joined_matchmake_session(self, client, input, output):
		logger.warning("DebugServer.get_exclude_joined_matchmake_session is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_api_call_summary(self, client, input, output):
		logger.warning("DebugServer.get_api_call_summary is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def enable_api_recorder(self, *args):
		logger.warning("DebugServer.enable_api_recorder not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def disable_api_recorder(self, *args):
		logger.warning("DebugServer.disable_api_recorder not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def is_api_recorder_enabled(self, *args):
		logger.warning("DebugServer.is_api_recorder_enabled not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_api_calls(self, *args):
		logger.warning("DebugServer.get_api_calls not implemented")
		raise common.RMCError("Core::NotImplemented")





# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class NATTraversalProtocol:
	METHOD_REQUEST_PROBE_INITIATION = 1
	METHOD_INITIATE_PROBE = 2
	METHOD_REQUEST_PROBE_INITIATION_EXT = 3
	METHOD_REPORT_NAT_TRAVERSAL_RESULT = 4
	METHOD_REPORT_NAT_PROPERTIES = 5
	METHOD_GET_RELAY_SIGNATURE_KEY = 6
	METHOD_REPORT_NAT_TRAVERSAL_RESULT_DETAIL = 7
	
	PROTOCOL_ID = 0x3


class NATTraversalClient(NATTraversalProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def request_probe_initiation(self, target_urls):
		logger.info("NATTraversalClient.request_probe_initiation()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(target_urls, stream.stationurl)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REQUEST_PROBE_INITIATION, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("NATTraversalClient.request_probe_initiation -> done")
	
	async def initiate_probe(self, station_to_probe):
		logger.info("NATTraversalClient.initiate_probe()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.stationurl(station_to_probe)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_INITIATE_PROBE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("NATTraversalClient.initiate_probe -> done")
	
	async def request_probe_initiation_ext(self, target_urls, station_to_probe):
		logger.info("NATTraversalClient.request_probe_initiation_ext()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(target_urls, stream.stationurl)
		stream.stationurl(station_to_probe)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REQUEST_PROBE_INITIATION_EXT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("NATTraversalClient.request_probe_initiation_ext -> done")
	
	async def report_nat_traversal_result(self, cid, result):
		logger.info("NATTraversalClient.report_nat_traversal_result()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(cid)
		stream.bool(result)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REPORT_NAT_TRAVERSAL_RESULT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("NATTraversalClient.report_nat_traversal_result -> done")
	
	async def report_nat_properties(self, natm, natf, rtt):
		logger.info("NATTraversalClient.report_nat_properties()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(natm)
		stream.u32(natf)
		stream.u32(rtt)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REPORT_NAT_PROPERTIES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("NATTraversalClient.report_nat_properties -> done")
	
	async def get_relay_signature_key(self):
		logger.info("NATTraversalClient.get_relay_signature_key()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RELAY_SIGNATURE_KEY, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.mode = stream.s32()
		obj.time = stream.datetime()
		obj.address = stream.string()
		obj.port = stream.u16()
		obj.address_type = stream.s32()
		obj.game_server_id = stream.u32()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("NATTraversalClient.get_relay_signature_key -> done")
		return obj
	
	async def report_nat_traversal_result_detail(self, cid, result, detail, rtt):
		logger.info("NATTraversalClient.report_nat_traversal_result_detail()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(cid)
		stream.bool(result)
		stream.s32(detail)
		stream.u32(rtt)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REPORT_NAT_TRAVERSAL_RESULT_DETAIL, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("NATTraversalClient.report_nat_traversal_result_detail -> done")


class NATTraversalServer(NATTraversalProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_REQUEST_PROBE_INITIATION: self.handle_request_probe_initiation,
			self.METHOD_INITIATE_PROBE: self.handle_initiate_probe,
			self.METHOD_REQUEST_PROBE_INITIATION_EXT: self.handle_request_probe_initiation_ext,
			self.METHOD_REPORT_NAT_TRAVERSAL_RESULT: self.handle_report_nat_traversal_result,
			self.METHOD_REPORT_NAT_PROPERTIES: self.handle_report_nat_properties,
			self.METHOD_GET_RELAY_SIGNATURE_KEY: self.handle_get_relay_signature_key,
			self.METHOD_REPORT_NAT_TRAVERSAL_RESULT_DETAIL: self.handle_report_nat_traversal_result_detail,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on NATTraversalServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_request_probe_initiation(self, client, input, output):
		logger.info("NATTraversalServer.request_probe_initiation()")
		#--- request ---
		target_urls = input.list(input.stationurl)
		await self.request_probe_initiation(client, target_urls)
	
	async def handle_initiate_probe(self, client, input, output):
		logger.info("NATTraversalServer.initiate_probe()")
		#--- request ---
		station_to_probe = input.stationurl()
		await self.initiate_probe(client, station_to_probe)
	
	async def handle_request_probe_initiation_ext(self, client, input, output):
		logger.info("NATTraversalServer.request_probe_initiation_ext()")
		#--- request ---
		target_urls = input.list(input.stationurl)
		station_to_probe = input.stationurl()
		await self.request_probe_initiation_ext(client, target_urls, station_to_probe)
	
	async def handle_report_nat_traversal_result(self, client, input, output):
		logger.info("NATTraversalServer.report_nat_traversal_result()")
		#--- request ---
		cid = input.u32()
		result = input.bool()
		await self.report_nat_traversal_result(client, cid, result)
	
	async def handle_report_nat_properties(self, client, input, output):
		logger.info("NATTraversalServer.report_nat_properties()")
		#--- request ---
		natm = input.u32()
		natf = input.u32()
		rtt = input.u32()
		await self.report_nat_properties(client, natm, natf, rtt)
	
	async def handle_get_relay_signature_key(self, client, input, output):
		logger.info("NATTraversalServer.get_relay_signature_key()")
		#--- request ---
		response = await self.get_relay_signature_key(client)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['mode', 'time', 'address', 'port', 'address_type', 'game_server_id']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.s32(response.mode)
		output.datetime(response.time)
		output.string(response.address)
		output.u16(response.port)
		output.s32(response.address_type)
		output.u32(response.game_server_id)
	
	async def handle_report_nat_traversal_result_detail(self, client, input, output):
		logger.info("NATTraversalServer.report_nat_traversal_result_detail()")
		#--- request ---
		cid = input.u32()
		result = input.bool()
		detail = input.s32()
		rtt = input.u32()
		await self.report_nat_traversal_result_detail(client, cid, result, detail, rtt)
	
	async def request_probe_initiation(self, *args):
		logger.warning("NATTraversalServer.request_probe_initiation not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def initiate_probe(self, *args):
		logger.warning("NATTraversalServer.initiate_probe not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def request_probe_initiation_ext(self, *args):
		logger.warning("NATTraversalServer.request_probe_initiation_ext not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def report_nat_traversal_result(self, *args):
		logger.warning("NATTraversalServer.report_nat_traversal_result not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def report_nat_properties(self, *args):
		logger.warning("NATTraversalServer.report_nat_properties not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_relay_signature_key(self, *args):
		logger.warning("NATTraversalServer.get_relay_signature_key not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def report_nat_traversal_result_detail(self, *args):
		logger.warning("NATTraversalServer.report_nat_traversal_result_detail not implemented")
		raise common.RMCError("Core::NotImplemented")





from anynet import tls, http
from nintendo import resources
import base64

import logging
logger = logging.getLogger(__name__)


USER_AGENT = {
	 900: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 901: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 910: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 920: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	1000: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1001: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1002: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1003: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1004: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1010: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1011: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1020: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1100: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1101: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1200: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1201: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1202: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1203: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1210: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1300: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.3.0.0; Add-on 13.3.0.0)",
	1310: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1320: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1321: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1400: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1410: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1411: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1412: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1500: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 15.3.0.0; Add-on 15.3.0.0)",
	1501: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 15.3.0.0; Add-on 15.3.0.0)",
	1600: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1601: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1602: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1603: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1610: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1700: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 17.5.0.0; Add-on 17.5.0.0)",
	1701: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 17.5.0.0; Add-on 17.5.0.0)",
	1800: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1801: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1810: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1900: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 19.3.0.0; Add-on 19.3.0.0)",
	1901: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 19.3.0.0; Add-on 19.3.0.0)",
	2000: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2001: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2010: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2011: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2015: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2020: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2030: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2040: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2050: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2100: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
	2101: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
	2110: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
	2120: "libcurl (nnFriends; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
}

LATEST_VERSION = 2120

LANGUAGES = [
	"en-US", "en-GB", "ja", "fr", "de", "es-419", "es", "it", "nl"
	"fr-CA", "pt", "ru", "zh-Hans", "zh-Hant", "ko", "pt-BR"
]


class FiveError(Exception):
	INVALID_PARAMETER = 2
	INVALID_REQUEST_URI = 3
	UNAUTHORIZED = 6
	RESOURCE_NOT_FOUND = 10
	APPLICATION_DATA_TOO_LARGE = 11
	
	def __init__(self, response):
		self.response = response
		
		self.code = int(response.json["error"]["code"])
		self.message = response.json["error"]["message"]
	
	def __str__(self):
		return self.message


class FiveClient:
	def __init__(self):
		self.request_callback = http.request
		
		ca = resources.certificate("Nintendo_CA_G3.der")
		self.context = tls.TLSContext()
		self.context.set_authority(ca)
		
		self.host = "app.lp1.five.nintendo.net"
		
		self.system_version = LATEST_VERSION
		self.user_agent = USER_AGENT[LATEST_VERSION]
	
	def set_request_callback(self, callback): self.request_callback = callback
	def set_context(self, context): self.context = context
	def set_host(self, host): self.host = host
	
	def set_system_version(self, version):
		if version not in USER_AGENT:
			raise ValueError("Unknown system version")
		self.system_version = version
		self.user_agent = USER_AGENT[version]
	
	async def request(self, req, access_token):
		req.headers["Host"] = self.host
		req.headers["User-Agent"] = self.user_agent
		req.headers["Accept"] = "*/*"
		
		if req.method != "GET":
			if req.json is not None:
				req.headers["Content-Type"] = "application/json"
				req.headers["Authorization"] = "Bearer " + access_token
				req.headers["Content-Length"] = 0
			elif req.form is not None:
				req.headers["Content-Type"] = "application/x-www-form-urlencoded"
				req.headers["Authorization"] = "Bearer " + access_token
				req.headers["Content-Length"] = 0
			else:
				req.headers["Authorization"] = "Bearer " + access_token
				req.headers["Content-Length"] = 0
				req.headers["Content-Type"] = "application/x-www-form-urlencoded"
		else:
			req.headers["Authorization"] = "Bearer " + access_token
		
		response = await self.request_callback(self.host, req, self.context)
		if response.json and "error" in response.json:
			logger.warning("Five server returned an error: %s" %response.json)
			raise FiveError(response)
		response.raise_if_error()
		return response
	
	async def get_unread_invitation_count(self, access_token, user_id):
		if self.system_version < 2000:
			req = http.HTTPRequest.get("/v1/users/%016x/invitations/inbox" %user_id)
			req.params = {
				"fields": "count",
				"read": "false"
			}
		else:
			req = http.HTTPRequest.get("/v2/users/%016x/invitations/inbox" %user_id)
			req.params = {
				"fields": "count",
				"read": "false",
				"invitation_types": "friend"
			}
		response = await self.request(req, access_token)
		return response.json["count"]
	
	async def get_inbox(self, access_token, user_id):
		if self.system_version < 2000:
			req = http.HTTPRequest.get("/v1/users/%016x/invitations/inbox" %user_id)
		else:
			req = http.HTTPRequest.get("/v2/users/%016x/invitations/inbox" %user_id)
			req.params = {
				"invitation_types": "friend"
			}
		response = await self.request(req, access_token)
		return response.json

	async def get_invitation_group(self, access_token, invitation_group_id):
		req = http.HTTPRequest.get("/v1/invitation_groups/%i" %invitation_group_id)
		response = await self.request(req, access_token)
		return response.json
	
	async def mark_as_read(self, access_token, ids):
		req = http.HTTPRequest.patch("/v1/invitations")
		req.form = {
			"read": "true",
			"ids": ",".join("%016x" %id for id in ids)
		}
		await self.request(req, access_token)
	
	async def mark_all_as_read(self, access_token, user_id):
		req = http.HTTPRequest.patch("/v1/users/%016x/invitations/mark_as_read" %user_id)
		await self.request(req, access_token)

	async def send_invitation(
		self, access_token, receivers, application_id, application_group_id,
		application_data, messages, application_id_match=False,
		acd_index=0
	):
		# Sanity checks
		if len(receivers) > 16:
			raise ValueError("Too many receiver ids")
		for language, message in messages.items():
			if language not in LANGUAGES:
				raise ValueError("'%s' is not a valid language" %language)
			if len(message) >= 0xC0:
				raise ValueError("Message for language '%s' is too long" %language)
		if len(application_data) > 0x400:
			raise ValueError("Application data is too large")
		
		url = "/v1/invitation_groups"
		if self.system_version >= 2000:
			url = "/v2/invitation_groups"
		
		req = http.HTTPRequest.post(url)

		req.json = {
			"receiver_ids": ["%016x" %id for id in receivers],
		}

		if self.system_version >= 2000:
			req.json["invitation_type"] = "friend"
		
		req.json["application_id"] = "%016x" %application_id

		if self.system_version >= 1900:
			req.json["acd_index"] = acd_index
		
		req.json["application_group_id"] = "%016x" %application_group_id
		
		if application_data:
			req.json["application_data"] = base64.b64encode(application_data).decode()
		
		req.json["messages"] = messages
		for language in LANGUAGES:
			if language in messages:
				req.json["messages"][language] = messages[language]
		req.json["application_id_match"] = application_id_match
		
		req.json_options["ensure_ascii"] = False
		
		response = await self.request(req, access_token)
		return response.json




from Crypto.Cipher import ARC4
from nintendo.nex import streams
import struct
import secrets
import hashlib
import hmac


class KeyDerivationOld:
	def __init__(self, base_count = 65000, pid_count = 1024):
		self.base_count = base_count
		self.pid_count = pid_count
		
	def derive_key(self, password, pid):
		key = password
		for i in range(self.base_count + pid % self.pid_count):
			key = hashlib.md5(key).digest()
		return key
		
		
class KeyDerivationNew:
	def __init__(self, base_count = 1, pid_count = 1):
		self.base_count = base_count
		self.pid_count = pid_count
		
	def derive_key(self, password, pid):
		key = password
		for i in range(self.base_count):
			key = hashlib.md5(key).digest()
			
		key += struct.pack("<Q", pid)
		for i in range(self.pid_count):
			key = hashlib.md5(key).digest()
			
		return key


class KerberosEncryption:
	def __init__(self, key):
		self.key = key
		
	def check(self, buffer):
		data = buffer[:-0x10]
		checksum = buffer[-0x10:]
		mac = hmac.new(self.key, data, digestmod=hashlib.md5)
		return checksum == mac.digest()
		
	def decrypt(self, buffer):
		if not self.check(buffer):
			raise ValueError("Invalid Kerberos checksum (incorrect password)")
		return ARC4.new(self.key).decrypt(buffer[:-0x10])
		
	def encrypt(self, buffer):
		encrypted = ARC4.new(self.key).encrypt(buffer)
		mac = hmac.new(self.key, encrypted, digestmod=hashlib.md5)
		return encrypted + mac.digest()


class ClientTicket:
	def __init__(self):
		self.session_key = None
		self.target = None
		self.internal = b""
	
	@classmethod
	def decrypt(cls, data, key, settings):
		kerberos = KerberosEncryption(key)
		decrypted = kerberos.decrypt(data)
		stream = streams.StreamIn(decrypted, settings)
		
		ticket = cls()
		ticket.session_key = stream.read(settings["kerberos.key_size"])
		ticket.target = stream.pid()
		ticket.internal = stream.buffer()
		return ticket

	def encrypt(self, key, settings):
		stream = streams.StreamOut(settings)
		if settings["kerberos.key_size"] != len(self.session_key):
			raise ValueError("Incorrect session_key size")
		stream.write(self.session_key)
		stream.pid(self.target)
		stream.buffer(self.internal)

		data = stream.get()
		kerberos = KerberosEncryption(key)
		return kerberos.encrypt(data)
		
		
class ServerTicket:
	def __init__(self):
		self.timestamp = None
		self.source = None
		self.session_key = None
	
	@classmethod
	def decrypt(cls, data, key, settings):
		if settings["kerberos.ticket_version"] == 1:
			stream = streams.StreamIn(data, settings)
			ticket_key = stream.buffer()
			data = stream.buffer()
			key = hashlib.md5(key + ticket_key).digest()

		kerberos = KerberosEncryption(key)
		decrypted = kerberos.decrypt(data)
		
		stream = streams.StreamIn(decrypted, settings)
		
		ticket = cls()
		ticket.timestamp = stream.datetime()
		ticket.source = stream.pid()
		ticket.session_key = stream.read(settings["kerberos.key_size"])
		return ticket

	def encrypt(self, key, settings):
		stream = streams.StreamOut(settings)
		stream.datetime(self.timestamp)
		stream.pid(self.source)
		if len(self.session_key) != settings["kerberos.key_size"]:
			raise ValueError("Incorrect session_key length")
		stream.write(self.session_key)

		data = stream.get()
		if settings["kerberos.ticket_version"] == 1:
			ticket_key = secrets.token_bytes(16)
			final_key = hashlib.md5(key + ticket_key).digest()

			kerberos = KerberosEncryption(final_key)
			encrypted = kerberos.encrypt(data)
			
			stream = streams.StreamOut(settings)
			stream.buffer(ticket_key)
			stream.buffer(encrypted)
			return stream.get()
		
		kerberos = KerberosEncryption(key)
		encrypted = kerberos.encrypt(data)
		return encrypted


class Credentials:
	def __init__(self, ticket, pid, cid):
		self.ticket = ticket
		self.pid = pid
		self.cid = cid




import setuptools

long_description = \
	"This library implements various network protocols made by Nintendo."

setuptools.setup(
	name = "nintendoclients",
	version = "4.3.0",
	description = "Nintendo network library",
	long_description = long_description,
	author = "Yannik Marchand",
	author_email = "ymarchand@me.com",
	url = "https://github.com/kinnay/NintendoClients",
	license = "MIT",
	packages = [
		"nintendo", "nintendo.nex", "nintendo.switch"
	],
	package_data = {
		"nintendo": ["files/config/*", "files/cert/*"]
	},
	install_requires = [
		"anynet ~= 1.1",
		"pycryptodome"
	]
)




error_names = {
	0x00010001: "Core::Unknown",
	0x00010002: "Core::NotImplemented",
	0x00010003: "Core::InvalidPointer",
	0x00010004: "Core::OperationAborted",
	0x00010005: "Core::Exception",
	0x00010006: "Core::AccessDenied",
	0x00010007: "Core::InvalidHandle",
	0x00010008: "Core::InvalidIndex",
	0x00010009: "Core::OutOfMemory",
	0x0001000A: "Core::InvalidArgument",
	0x0001000B: "Core::Timeout",
	0x0001000C: "Core::InitializationFailure",
	0x0001000D: "Core::CallInitiationFailure",
	0x0001000E: "Core::RegistrationError",
	0x0001000F: "Core::BufferOverflow",
	0x00010010: "Core::InvalidLockState",
	0x00010011: "Core::InvalidSequence",
	0x00010012: "Core::SystemError",
	0x00010013: "Core::Cancelled",
	
	0x00020001: "DDL::InvalidSignature",
	0x00020002: "DDL::IncorrectVersion",
	
	0x00030001: "RendezVous::ConnectionFailure",
	0x00030002: "RendezVous::NotAuthenticated",

	0x00030064: "RendezVous::InvalidUsername",
	0x00030065: "RendezVous::InvalidPassword",
	0x00030066: "RendezVous::UsernameAlreadyExists",
	0x00030067: "RendezVous::AccountDisabled",
	0x00030068: "RendezVous::AccountExpired",
	0x00030069: "RendezVous::ConcurrentLoginDenied",
	0x0003006A: "RendezVous::EncryptionFailure",
	0x0003006B: "RendezVous::InvalidPID",
	0x0003006C: "RendezVous::MaxConnectionsReached",
	0x0003006D: "RendezVous::InvalidGID",
	0x0003006E: "RendezVous::InvalidControlScriptID",
	0x0003006F: "RendezVous::InvalidOperationInLiveEnvironment",
	0x00030070: "RendezVous::DuplicateEntry",
	0x00030071: "RendezVous::ControlScriptFailure",
	0x00030072: "RendezVous::ClassNotFound",
	0x00030073: "RendezVous::SessionVoid",
	0x00030075: "RendezVous::DDLMismatch",
	0x00030076: "RendezVous::InvalidConfiguration",
	
	0x000300C8: "RendezVous::SessionFull",
	0x000300C9: "RendezVous::InvalidGatheringPassword",
	0x000300CA: "RendezVous::WithoutParticipationPeriod",
	0x000300CB: "RendezVous::PersistentGatheringCreationMax",
	0x000300CC: "RendezVous::PersistentGatheringParticipationMax",
	0x000300CD: "RendezVous::DeniedByParticipants",
	0x000300CE: "RendezVous::ParticipantInBlackList",
	0x000300CF: "RendezVous::GameServerMaintenance",
	0x000300D0: "RendezVous::OperationPostpone",
	0x000300D1: "RendezVous::OutOfRatingRange",
	0x000300D2: "RendezVous::ConnectionDisconnected",
	0x000300D3: "RendezVous::InvalidOperation",
	0x000300D4: "RendezVous::NotParticipatedGathering",
	0x000300D5: "RendezVous::MatchmakeSessionUserPasswordUnmatch",
	0x000300D6: "RendezVous::MatchmakeSessionSystemPasswordUnmatch",
	0x000300D7: "RendezVous::UserIsOffline",
	0x000300D8: "RendezVous::AlreadyParticipatedGathering",
	0x000300D9: "RendezVous::PermissionDenied",
	0x000300DA: "RendezVous::NotFriend",
	0x000300DB: "RendezVous::SessionClosed",
	0x000300DC: "RendezVous::DatabaseTemporarilyUnavailable",
	0x000300DD: "RendezVous::InvalidUniqueId",
	0x000300DE: "RendezVous::MatchmakingWithdrawn",
	0x000300DF: "RendezVous::LimitExceeded",
	0x000300E0: "RendezVous::AccountTemporarilyDisabled",
	0x000300E1: "RendezVous::PartiallyServiceClosed",
	0x000300E2: "RendezVous::ConnectionDisconnectedForConcurrentLogin",
	
	0x00040001: "PythonCore::Exception",
	0x00040002: "PythonCore::TypeError",
	0x00040003: "PythonCore::IndexError",
	0x00040004: "PythonCore::InvalidReference",
	0x00040005: "PythonCore::CallFailure",
	0x00040006: "PythonCore::MemoryError",
	0x00040007: "PythonCore::KeyError",
	0x00040008: "PythonCore::OperationError",
	0x00040009: "PythonCore::ConversionError",
	0x0004000A: "PythonCore::ValidationError",
	
	0x00050001: "Transport::Unknown",
	0x00050002: "Transport::ConnectionFailure",
	0x00050003: "Transport::InvalidUrl",
	0x00050004: "Transport::InvalidKey",
	0x00050005: "Transport::InvalidURLType",
	0x00050006: "Transport::DuplicateEndpoint",
	0x00050007: "Transport::IOError",
	0x00050008: "Transport::Timeout",
	0x00050009: "Transport::ConnectionReset",
	0x0005000A: "Transport::IncorrectRemoteAuthentication",
	0x0005000B: "Transport::ServerRequestError",
	0x0005000C: "Transport::DecompressionFailure",
	0x0005000D: "Transport::ReliableSendBufferFullFatal",
	0x0005000E: "Transport::UPnPCannotInit",
	0x0005000F: "Transport::UPnPCannotAddMapping",
	0x00050010: "Transport::NatPMPCannotInit",
	0x00050011: "Transport::NatPMPCannotAddMapping",
	0x00050013: "Transport::UnsupportedNAT",
	0x00050014: "Transport::DnsError",
	0x00050015: "Transport::ProxyError",
	0x00050016: "Transport::DataRemaining",
	0x00050017: "Transport::NoBuffer",
	0x00050018: "Transport::NotFound",
	0x00050019: "Transport::TemporaryServerError",
	0x0005001A: "Transport::PermanentServerError",
	0x0005001B: "Transport::ServiceUnavailable",
	0x0005001C: "Transport::ReliableSendBufferFull",
	0x0005001D: "Transport::InvalidStation",
	0x0005001E: "Transport::InvalidSubStreamID",
	0x0005001F: "Transport::PacketBufferFull",
	0x00050020: "Transport::NatTraversalError",
	0x00050021: "Transport::NatCheckError",
	
	0x00060001: "DOCore::StationNotReached",
	0x00060002: "DOCore::TargetStationDisconnect",
	0x00060003: "DOCore::LocalStationLeaving",
	0x00060004: "DOCore::ObjectNotFound",
	0x00060005: "DOCore::InvalidRole",
	0x00060006: "DOCore::CallTimeout",
	0x00060007: "DOCore::RMCDispatchFailed",
	0x00060008: "DOCore::MigrationInProgress",
	0x00060009: "DOCore::NoAuthority",
	0x0006000A: "DOCore::NoTargetStationSpecified",
	0x0006000B: "DOCore::JoinFailed",
	0x0006000C: "DOCore::JoinDenied",
	0x0006000D: "DOCore::ConnectivityTestFailed",
	0x0006000E: "DOCore::Unknown",
	0x0006000F: "DOCore::UnfreedReferences",
	0x00060010: "DOCore::JobTerminationFailed",
	0x00060011: "DOCore::InvalidState",
	0x00060012: "DOCore::FaultRecoveryFatal",
	0x00060013: "DOCore::FaultRecoveryJobProcessFailed",
	0x00060014: "DOCore::StationInconsitency",
	0x00060015: "DOCore::AbnormalMasterState",
	0x00060016: "DOCore::VersionMismatch",
	
	0x00650000: "FPD::NotInitialized",
	0x00650001: "FPD::AlreadyInitialized",
	0x00650002: "FPD::NotConnected",
	0x00650003: "FPD::Connected",
	0x00650004: "FPD::InitializationFailure",
	0x00650005: "FPD::OutOfMemory",
	0x00650006: "FPD::RmcFailed",
	0x00650007: "FPD::InvalidArgument",
	0x00650008: "FPD::InvalidLocalAccountID",
	0x00650009: "FPD::InvalidPrincipalID",
	0x0065000A: "FPD::InvalidLocalFriendCode",
	0x0065000B: "FPD::LocalAccountNotExists",
	0x0065000C: "FPD::LocalAccountNotLoaded",
	0x0065000D: "FPD::LocalAccountAlreadyLoaded",
	0x0065000E: "FPD::FriendAlreadyExists",
	0x0065000F: "FPD::FriendNotExists",
	0x00650010: "FPD::FriendNumMax",
	0x00650011: "FPD::NotFriend",
	0x00650012: "FPD::FileIO",
	0x00650013: "FPD::P2PInternetProhibited",
	0x00650014: "FPD::Unknown",
	0x00650015: "FPD::InvalidState",
	0x00650017: "FPD::AddFriendProhibited",
	0x00650019: "FPD::InvalidAccount",
	0x0065001A: "FPD::BlacklistedByMe",
	0x0065001C: "FPD::FriendAlreadyAdded",
	0x0065001D: "FPD::MyFriendListLimitExceed",
	0x0065001E: "FPD::RequestLimitExceed",
	0x0065001F: "FPD::InvalidMessageID",
	0x00650020: "FPD::MessageIsNotMine",
	0x00650021: "FPD::MessageIsNotForMe",
	0x00650022: "FPD::FriendRequestBlocked",
	0x00650023: "FPD::NotInMyFriendList",
	0x00650024: "FPD::FriendListedByMe",
	0x00650025: "FPD::NotInMyBlacklist",
	0x00650026: "FPD::IncompatibleAccount",
	0x00650027: "FPD::BlockSettingChangeNotAllowed",
	0x00650028: "FPD::SizeLimitExceeded",
	0x00650029: "FPD::OperationNotAllowed",
	0x0065002A: "FPD::NotNetworkAccount",
	0x0065002B: "FPD::NotificationNotFound",
	0x0065002C: "FPD::PreferenceNotInitialized",
	0x0065002D: "FPD::FriendRequestNotAllowed",
	
	0x00670001: "Ranking::NotInitialized",
	0x00670002: "Ranking::InvalidArgument",
	0x00670003: "Ranking::RegistrationError",
	0x00670005: "Ranking::NotFound",
	0x00670006: "Ranking::InvalidScore",
	0x00670007: "Ranking::InvalidDataSize",
	0x00670009: "Ranking::PermissionDenied",
	0x0067000A: "Ranking::Unknown",
	0x0067000B: "Ranking::NotImplemented",
	
	0x00680001: "Authentication::NASAuthenticateError",
	0x00680002: "Authentication::TokenParseError",
	0x00680003: "Authentication::HttpConnectionError",
	0x00680004: "Authentication::HttpDNSError",
	0x00680005: "Authentication::HttpGetProxySetting",
	0x00680006: "Authentication::TokenExpired",
	0x00680007: "Authentication::ValidationFailed",
	0x00680008: "Authentication::InvalidParam",
	0x00680009: "Authentication::PrincipalIdUnmatched",
	0x0068000A: "Authentication::MoveCountUnmatch",
	0x0068000B: "Authentication::UnderMaintenance",
	0x0068000C: "Authentication::UnsupportedVersion",
	0x0068000D: "Authentication::ServerVersionIsOld",
	0x0068000E: "Authentication::Unknown",
	0x0068000F: "Authentication::ClientVersionIsOld",
	0x00680010: "Authentication::AccountLibraryError",
	0x00680011: "Authentication::ServiceNoLongerAvailable",
	0x00680012: "Authentication::UnknownApplication",
	0x00680013: "Authentication::ApplicationVersionIsOld",
	0x00680014: "Authentication::OutOfService",
	0x00680015: "Authentication::NetworkServiceLicenseRequired",
	0x00680016: "Authentication::NetworkServiceLicenseSystemError",
	0x00680017: "Authentication::NetworkServiceLicenseError3",
	0x00680018: "Authentication::NetworkServiceLicenseError4",
	
	0x00690001: "DataStore::Unknown",
	0x00690002: "DataStore::InvalidArgument",
	0x00690003: "DataStore::PermissionDenied",
	0x00690004: "DataStore::NotFound",
	0x00690005: "DataStore::AlreadyLocked",
	0x00690006: "DataStore::UnderReviewing",
	0x00690007: "DataStore::Expired",
	0x00690008: "DataStore::InvalidCheckToken",
	0x00690009: "DataStore::SystemFileError",
	0x0069000A: "DataStore::OverCapacity",
	0x0069000B: "DataStore::OperationNotAllowed",
	0x0069000C: "DataStore::InvalidPassword",
	0x0069000D: "DataStore::ValueNotEqual",
	
	0x006C0001: "ServiceItem::Unknown",
	0x006C0002: "ServiceItem::InvalidArgument",
	0x006C0003: "ServiceItem::EShopUnknownHttpError",
	0x006C0004: "ServiceItem::EShopResponseParseError",
	0x006C0005: "ServiceItem::NotOwned",
	0x006C0006: "ServiceItem::InvalidLimitationType",
	0x006C0007: "ServiceItem::ConsumptionRightShortage",
	
	0x006F0001: "MatchmakeReferee::Unknown",
	0x006F0002: "MatchmakeReferee::InvalidArgument",
	0x006F0003: "MatchmakeReferee::AlreadyExists",
	0x006F0004: "MatchmakeReferee::NotParticipatedGathering",
	0x006F0005: "MatchmakeReferee::NotParticipatedRound",
	0x006F0006: "MatchmakeReferee::StatsNotFound",
	0x006F0007: "MatchmakeReferee::RoundNotFound",
	0x006F0008: "MatchmakeReferee::RoundArbitrated",
	0x006F0009: "MatchmakeReferee::RoundNotArbitrated",
	
	0x00700001: "Subscriber::Unknown",
	0x00700002: "Subscriber::InvalidArgument",
	0x00700003: "Subscriber::OverLimit",
	0x00700004: "Subscriber::PermissionDenied",
	
	0x00710001: "Ranking2::Unknown",
	0x00710002: "Ranking2::InvalidArgument",
	0x00710003: "Ranking2::InvalidScore",
	
	0x00720001: "SmartDeviceVoiceChat::Unknown",
	0x00720002: "SmartDeviceVoiceChat::InvalidArgument",
	0x00720003: "SmartDeviceVoiceChat::InvalidResponse",
	0x00720004: "SmartDeviceVoiceChat::InvalidAccessToken",
	0x00720005: "SmartDeviceVoiceChat::Unauthorized",
	0x00720006: "SmartDeviceVoiceChat::AccessError",
	0x00720007: "SmartDeviceVoiceChat::UserNotFound",
	0x00720008: "SmartDeviceVoiceChat::RoomNotFound",
	0x00720009: "SmartDeviceVoiceChat::RoomNotActivated",
	0x0072000A: "SmartDeviceVoiceChat::ApplicationNotSupported",
	0x0072000B: "SmartDeviceVoiceChat::InternalServerError",
	0x0072000C: "SmartDeviceVoiceChat::ServiceUnavailable",
	0x0072000D: "SmartDeviceVoiceChat::UnexpectedError",
	0x0072000E: "SmartDeviceVoiceChat::UnderMaintenance",
	0x0072000F: "SmartDeviceVoiceChat::ServiceNoLongerAvailable",
	0x00720010: "SmartDeviceVoiceChat::AccountTemporarilyDisabled",
	0x00720011: "SmartDeviceVoiceChat::PermissionDenied",
	0x00720012: "SmartDeviceVoiceChat::NetworkServiceLicenseRequired",
	0x00720013: "SmartDeviceVoiceChat::AccountLibraryError",
	0x00720014: "SmartDeviceVoiceChat::GameModeNotFound",
	
	0x00730001: "Screening::Unknown",
	0x00730002: "Screening::InvalidArgument",
	0x00730003: "Screening::NotFound",
	
	0x00740001: "Custom::Unknown",

	0x00750001: "Ess::Unknown",
	0x00750002: "Ess::GameSessionError",
	0x00750003: "Ess::GameSessionMaintenance"
}

error_codes = {name: code for code, name in error_names.items()}




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class HealthProtocol:
	METHOD_PING_DAEMON = 1
	METHOD_PING_DATABASE = 2
	METHOD_RUN_SANITY_CHECK = 3
	METHOD_FIX_SANITY_ERRORS = 4
	
	PROTOCOL_ID = 0x12


class HealthClient(HealthProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def ping_daemon(self):
		logger.info("HealthClient.ping_daemon()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_PING_DAEMON, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("HealthClient.ping_daemon -> done")
		return result
	
	async def ping_database(self):
		logger.info("HealthClient.ping_database()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_PING_DATABASE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("HealthClient.ping_database -> done")
		return result
	
	async def run_sanity_check(self):
		logger.info("HealthClient.run_sanity_check()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_RUN_SANITY_CHECK, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("HealthClient.run_sanity_check -> done")
		return result
	
	async def fix_sanity_errors(self):
		logger.info("HealthClient.fix_sanity_errors()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_FIX_SANITY_ERRORS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("HealthClient.fix_sanity_errors -> done")
		return result


class HealthServer(HealthProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_PING_DAEMON: self.handle_ping_daemon,
			self.METHOD_PING_DATABASE: self.handle_ping_database,
			self.METHOD_RUN_SANITY_CHECK: self.handle_run_sanity_check,
			self.METHOD_FIX_SANITY_ERRORS: self.handle_fix_sanity_errors,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on HealthServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_ping_daemon(self, client, input, output):
		logger.info("HealthServer.ping_daemon()")
		#--- request ---
		response = await self.ping_daemon(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_ping_database(self, client, input, output):
		logger.info("HealthServer.ping_database()")
		#--- request ---
		response = await self.ping_database(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_run_sanity_check(self, client, input, output):
		logger.info("HealthServer.run_sanity_check()")
		#--- request ---
		response = await self.run_sanity_check(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_fix_sanity_errors(self, client, input, output):
		logger.info("HealthServer.fix_sanity_errors()")
		#--- request ---
		response = await self.fix_sanity_errors(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def ping_daemon(self, *args):
		logger.warning("HealthServer.ping_daemon not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def ping_database(self, *args):
		logger.warning("HealthServer.ping_database not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def run_sanity_check(self, *args):
		logger.warning("HealthServer.run_sanity_check not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def fix_sanity_errors(self, *args):
		logger.warning("HealthServer.fix_sanity_errors not implemented")
		raise common.RMCError("Core::NotImplemented")

